/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"
#include "support.h"

#include <alloca.h>
#ifdef USECURSES
#include <curses.h>
#else
#include <ncurses.h>
#endif
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <term.h>
#include <unistd.h>


static WINDOW *w = 0, *H = 0;
enum { BLACK_ON_WHITE = 1, BLUE_ON_WHITE, RED_ON_WHITE };
static int Sigwinch = 1, stress, nostress, header, underline, 
    Numlines = 0, Numcol = 0, Oldcurs;
static sysstat_t *displayed = 0;
static clock_t Dis_ts;
extern int Fslines;
static const char *ti_nostress, *ti_stress, *ti_header, *ti_underline;

void
exit_curses(void)
{
	curs_set(Oldcurs);
	endwin();
}


void
sighandler(int sig)
{
	if (sig == SIGINT || sig == SIGTERM) {
		exit(EXIT_SUCCESS);
	}
	assert(sig == SIGWINCH);
	Sigwinch = 1;
}


void
display_help(void)
{
	char *line;

	if (H == 0) {
		if ((Numlines == 0) || (Numcol == 0)) {
			struct winsize w;
			if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&w))
				error("cannot read terminal size: %s\n",strerror(errno));
			Numlines = w.ws_row;
			Numcol = w.ws_col;
		}
		assert(Numlines && Numcol);
		H = newwin(Numlines,Numcol,0,0);
		line = alloca(Numcol + 1);
		wclear(H);
		memset(line,'=',Numcol);
		line[Numcol] = 0;
		mvwaddstr(H, 0,0,
			"sysstat version "VERSION"\n"
			"Copyright 2006, Thomas Maier-Komor\n"
			"license: GPLv2 - see file COPYING or \"man gpl\" for details\n\n"
			
			"Hotkeys:                         Field descriptions:\n"
			"--------                         -------------------\n"
			"'q' : quit sysstat               LoadAv:    load average over time interval\n"
			"'?' : toggle help screen         intr:      interrupts\n"
			"                                 intrblk:   blocked interrupts\n"
			"'h' : display next host          xcalls:    cpu to cpu cross calls\n"
			"'H' : display previous host      pswitch:   process/context switches\n"
			"                                 migrate:   inter cpu migrations of processes\n"
			"sort process list by:            invswtch:  involuntary context switches\n"
			"'u' : username                   scan:      pager deamon's page scans\n"
			"'n' : NPROC                      pgswap:    pages swapped in/out\n"
			"'r' : RSS                        anonpg:    anonymous pages paged in/out\n"
			"'s' : SIZE                       anonfree:  anonymous pages freed\n"
			"'+' : H+S                        NPROC:     number of processes\n"
			"'t' : Time                       O/S/R/Z/T: oncpu/sleep/runnable/zompie/stopped\n"
			"'c' : %CPU                       NLWP:      number of light weight processes\n"
			"                                 SIZE:      allocated memory\n"
			"                                 RSS:       resident memory\n"
			"                                 H+S:       heap plus stack\n"
			);
		wsyncup(H);
		wrefresh(H);
	} else {
		delwin(H);
		H = 0;
		refresh();
	}
}


int
init_curses(void)
{
	void (*osig)(int);

	if (0 == isatty(STDOUT_FILENO))
		return 0;
	w = initscr();
	start_color();
	noecho();
	cbreak();
	intrflush(stdscr,FALSE);
	nodelay(w, TRUE);
	nonl();
	Oldcurs = curs_set(0);
	if (has_colors()) {
		dbug("terminal has colors\n");
		init_pair(BLACK_ON_WHITE,COLOR_BLACK,COLOR_WHITE);
		init_pair(BLUE_ON_WHITE,COLOR_BLUE,COLOR_WHITE);
		init_pair(RED_ON_WHITE,COLOR_RED,COLOR_WHITE);
		nostress = COLOR_PAIR(BLACK_ON_WHITE);
		stress = A_BOLD | COLOR_PAIR(RED_ON_WHITE);
		header = COLOR_PAIR(BLUE_ON_WHITE) | A_UNDERLINE;
		underline = COLOR_PAIR(BLACK_ON_WHITE) | A_UNDERLINE;
	} else {
		dbug("terminal has no colors\n");
		header = A_UNDERLINE;
		underline = A_UNDERLINE;
		nostress = 0;
		stress = A_BOLD;
	}
	osig = signal(SIGWINCH,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGINT,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGTERM,sighandler);
	assert(osig != SIG_ERR);
	atexit(exit_curses);
	return 1;
}


unsigned 
display_curses(sysstat_t *s)
{
	int num;
	userstat_t *u;
	fsstat_t *fs;
	char line[80*25+1], *l, *const e = line + sizeof(line);
	int fslines = Fslines, ulines = 0, unshown = 0;
	sysstat_t *h = s;
	clock_t now;

	now = time(0);
	if (H || (s == 0) || (s->TS == 0) ||
	    ((s == displayed) && (s->TS <= Dis_ts) && (now - Dis_ts < 60) && (Sigwinch == 0)))
		return 0;
	displayed = s;
	Dis_ts = s->TS;
	h = s->next;
	while (h != s) {
		++unshown;
		h = h->next;
	}
	assert(w);
	if (Sigwinch) {
		struct winsize ws;
		if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&ws))
			error("cannot read terminal size: %s\n",strerror(errno));
		LINES = Numlines = ws.ws_row;
		COLS = Numcol = ws.ws_col;
		assert(COLS < sizeof(line));
		dbug("SIGWINCH: Numlines = %d\n",Numlines);
		endwin();
		refresh();
		redrawwin(stdscr);
#ifdef USECURSES
		w = initscr();
		start_color();
#endif
		Sigwinch = 0;
		(void) signal(SIGWINCH, sighandler);
	}
	if (unshown) 
		--Numlines;
	num = 0;
	if (fslines > s->numfs) {
		dbug("fslines = %d, numfs = %d\n",fslines,s->numfs);
		fslines = s->numfs;
	}
	attrset(nostress);
	erase();
	attrset(header);
	memset(line,' ',sizeof(line));
	l = line + strlcpy(line,s->Hostinfo,sizeof(line));
	if (s->Ocpus == s->Ncpus) {
		if (s->Ncpus > 1) {
			l += sprintf(l,", %d CPUs", s->Ocpus);
		} else {
			assert(s->Ocpus == 1);
			l += sprintf(l,", 1 CPU");
		}
	} else {
		assert(s->Ncpus > 1);
		l += sprintf(l,", %d/%d CPUs",s->Ocpus,s->Ncpus);
	}
	if (s->BootTime) {
		int d,h,m;
		clock_t b = s->TS - s->BootTime;
		d = b / (60*60*24);
		b -= d * (60*60*24);
		h = b / (60*60);
		b -= h * (60*60);
		m = b / (60);
		b -= m * (60);
		if (d == 1)
			l += sprintf(l,", up 1 day");
		else if (d)
			l += sprintf(l,", up %d days",d);
		else
			l += sprintf(l,", up %d:%02d:%02d",h,m,b);
	} else if (s->Devices != 1) {
		l += sprintf(l,", %lu disks ", s->Devices);
	} else if (s->Devices == 1) {
		l += sprintf(l,", 1 disk ");
	}
	*l = ' ';
	strftime(line+59,23,"  %c",localtime(&s->TS));
	if (now - s->TS > 60) {
		Dis_ts = now;
		attrset(stress);
	}
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(nostress);
	l = memccpy(line," User   CPU: ",0,sizeof(line));
	l = Per2str(s->UserP, --l);
	l = memccpy(l,"% | LoadAv/1m:  ",0,e - l);
	l = Dbl2str(s->loadavg[0],--l);
	l = memccpy(l," | Physical:    ",0,e - l);
	l = Kb2str(s->Physmem,--l);
	if (s->IpInR >= 0) {
		l = memccpy(l," | IP rcv/s:   ",0,e - l);
		l = Dbl2str(s->IpInR,--l);
	} else {
		l = memccpy(l," | ipacket/s:  ",0,e - l);
		l = Dbl2str(s->NetInR,--l);
	}
	*l = ' ';
	++l;
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = memccpy(line," Kernel CPU: ",0,sizeof(line));
	l = Per2str(s->KernelP,--l);
	l = memccpy(l,"% | LoadAv/5m:  ",0,e - l);
	l = Dbl2str(s->loadavg[1],--l);
	l = memccpy(l," | Free:        ",0,e - l);
	l = Kb2str(s->Pagesfree,--l);
	if (s->IpOutR >= 0) {
		l = memccpy(l," | IP send/s:  ",0,e - l);
		l = Dbl2str(s->IpOutR,--l);
	} else {
		l = memccpy(l," | opacket/s:  ",0,e - l);
		l = Dbl2str(s->NetOutR,--l);
	}
	*l = ' ';
	++l;
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(underline);
	mvaddstr(num++,0," Idle   CPU: ");
	if (s->IdleP <= 1)
		attrset(stress | A_UNDERLINE);
	addstr(per2str(s->IdleP));
	attrset(underline);
	l = memccpy(line,"% | LoadAv/15m: ",0,e - line);
	l = Dbl2str(s->loadavg[2],--l);
	l = memccpy(l," | Locked:      ",0,e - l);
	l = Kb2str(s->Pageslocked,--l);
	l = memccpy(l," | net errs: ",0,e - l);
	sprintf(--l,"%6lu  ", s->IpInDiscards+s->IpOutDiscards+s->IpInErrs);
	addstr(line);
	if (num == Numlines) goto done;

	attrset(nostress);
	l = memccpy(line," intr/s:     ",0,e-line);
	l = Dbl2str(s->IntrR,--l);
	l = memccpy(l," | syscall/s:  ",0,e-l);
	l = Dbl2str(s->SyscallR,--l);
	l = memccpy(l," | SwapTotal:   ",0,e-l);
	l = Kb2str(s->SwapTotal,--l);
	l = memccpy(l," | nfsd ops/s: ",0,e-l);
	l = Dbl2str(s->NfsdR,--l);
	*l++ = ' ';
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = memccpy(line," intrblk/s:  ",0,e-line);
	l = Dbl2str(s->IntrBlkR,--l);
	l = memccpy(l," | sysfork/s:  ",0,e-l);
	l = Dbl2str(s->SysforkR,--l);
	l = memccpy(l," | SwapFree:    ",0,e-l);
	l = Kb2str(s->SwapFree,--l);
	l = memccpy(l," | NfsIn/s:    ",0,e-l);
	l = Kb2str(s->NfsIR,--l);
	*l++ = ' ';
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = memccpy(line," xcalls/s:   ",0,e-line);
	l = Dbl2str(s->XcallsR,--l);
	l = memccpy(l," | sysexec/s:  ",0,e-l);
	l = Dbl2str(s->SysexecR,--l);
	memccpy(l," | scan/s:      ",0,e-l);
	mvaddstr(num++,0,line);
	if (s->ScanR)
		attrset(stress);
	addstr(dbl2str(s->ScanR));
	if (s->ScanR)
		attrset(nostress);
	l = memccpy(line," | NfsOut/s:   ",0,sizeof(line));
	l = Kb2str(s->NfsOR,--l);
	*l++ = ' ';
	*l = 0;
	addstr(line);
	if (num == Numlines) goto done;

	l = memccpy(line," pswitch/s:  ",0,e-line);
	l = Dbl2str(s->PswitchR,--l);
	l = memccpy(l," | sysread/s:  ",0,e-l);
	l = Dbl2str(s->SysreadR,--l);
	if ((s->PgswapinR != 0) && (s->AnonpginR == 0)) {
		l = memccpy(l," | pgswapin/s:  ",0,e-l);
		l = Dbl2str(s->PgswapinR,--l);
	} else {
		l = memccpy(l," | anonpgin/s:  ",0,e-l);
		l = Dbl2str(s->AnonpginR,--l);
	}
	l = memccpy(l," | DiskIn/s:   ",0,e-l);
	l = Kb2str(s->DiskIR,--l);
	*l++ = ' ';
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	l = memccpy(line," migrate/s:  ",0,e-line);
	l = Dbl2str(s->MigrateR,--l);
	l = memccpy(l," | syswrite/s: ",0,e-l);
	l = Dbl2str(s->SyswriteR,--l);
	if ((s->PgswapoutR != 0) && (s->AnonpgoutR == 0)) {
		l = memccpy(l," | pgswapout/s: ",0,e-l);
		l = Dbl2str(s->PgswapoutR,--l);
	} else {
		l = memccpy(l," | anonpgout/s: ",0,e-l);
		l = Dbl2str(s->AnonpgoutR,--l);
	}
	l = memccpy(l," | DiskOut/s:  ",0,e-l);
	l = Kb2str(s->DiskOR,--l);
	*l++ = ' ';
	*l = 0;
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(underline);
	l = memccpy(line," invswtch/s: ",0,e-line);
	l = Dbl2str(s->InvSwtchR,--l);
	l = memccpy(l," |                   | anonfree/s:  ",0,e-l);
	l = Dbl2str(s->AnonfreeR,--l);
	l = memccpy(l," | disk errs: ",0,e-l);
	sprintf(--l,"%5lu  ", s->HardErrors+s->SoftErrors);
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	ulines = Numlines - num - fslines;
	if (ulines < fslines) {
		fslines = (Numlines - num) / 2;
		ulines = Numlines - num - fslines;
		assert(ulines >= fslines);
	}
	dbug("fslines = %d\n",fslines);
	attrset(nostress);
	s->us = sort_users(s->us);
	u = s->us;
	if ((num + 1) < (Numlines - fslines)) {
		unsigned numuser = 0;
		userstat_t *us = s->us;
		while (us) {
			++numuser;
			us = us->next;
		}
		l = line + sprintf(line,"%d",numuser);
		if (u->h_s) {
			if (numuser < 10)
				strcpy(l," User   NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %MEM  %CPU   CPU-TIME");
			else if (numuser < 100)
				strcpy(l," User  NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %MEM  %CPU   CPU-TIME");
			else 
				strcpy(l,"d User NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS    H+S   %MEM  %CPU   CPU-TIME");
		} else {
			if (numuser < 10)
				strcpy(l," User         NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %MEM  %CPU    CPU-TIME");
			else if (numuser < 100)
				strcpy(l," User        NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %MEM  %CPU    CPU-TIME");
			else 
				strcpy(l," User       NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE   RSS   %MEM  %CPU    CPU-TIME");
		}
		mvaddstr(num++,0,line);
	}
	while (u && (num < (Numlines-fslines))) {
		long hr,mn,sc;
		unsigned ul = 0;
		char *l = line;
		double pctmem = (double)u->rss/(double)s->Physmem*100;
		if ((u->next == 0) || (num == (Numlines - fslines - 1))) {
			ul = A_UNDERLINE;
			attrset(nostress | A_UNDERLINE);
		}
		assert(u->time.tv_sec >= 0);
		assert(u->time.tv_nsec >= 0);
		sc = u->time.tv_sec + (long) rint((double)u->time.tv_nsec/1000000000);
		mn = u->time.tv_sec / 60;
		sc %= 60;
		hr = mn / 60;
		mn -= hr * 60;
		assert((sc >= 0) && (sc < 60) && (mn >= 0) && (mn < 60) && (hr >= 0));
		if (u->h_s) 
			l += sprintf(line,"%-9s ", u->username);
		else
			l += sprintf(line,"%-15s ", u->username);
		l += sprintf(l,"%4d (%2d/%2d/", u->nproc, u->oncpu, u->sleeping);
		mvaddstr(num,0,line);
		l = line + sprintf(line,"%2d", u->runnable);
		if (u->runnable) {
			attrset(stress | ul);
			addstr(line);
			attrset(nostress | ul);
			l = line;
		}
		l += sprintf(l,"/%2d/%2d) %4d  ",
				u->zombie,
				u->stopped,
				u->nlwp);
		l = Kb2str(u->size,l);
		memcpy(l,"  ",2);
		l += 2;
		l = Kb2str(u->rss,l);
		memcpy(l,"  ",2);
		l += 2;
		if (u->h_s) {
			l = Kb2str(u->h_s,l);
			*l++ = ' ';
			*l++ = ' ';
		}
		if (u->pctmem < 9.995)
			l += sprintf(l,"%3.1f%%  ",pctmem);
		else
			l += sprintf(l,"%3.f%%  ",pctmem);
		if (u->pctcpu < 9.995)
			l += sprintf(l,"%3.1f%% ",u->pctcpu);
		else
			l += sprintf(l,"%3.f%% ",u->pctcpu);
		if (u->h_s)
			sprintf(l,"%4d:%02d:%02d",hr,mn,sc);
		else
			sprintf(l,"%5d:%02d:%02d ",hr,mn,sc);
		addstr(line);
		u = u->next;
		dbug("us@%d\n",num);
		++num;
	}
	fs = s->fs;
	attrset(nostress);
	while ((num < Numlines) && fs) {
		char *l = line;
		if (unshown) {
			if ((num + 1 == Numlines) || (0 == fs->next))
				attrset(underline);
		}
		if (strlen(fs->dev) > 30)
			l += sprintf(l,"%.27s... %-5s ",fs->dev,fs->typ);
		else
			l += sprintf(l,"%-30s %-5s ",fs->dev,fs->typ);
		if (strlen(fs->mnt) > 30)
			l += sprintf(l,"%.27s...  ",fs->mnt);
		else
			l += sprintf(l,"%-30s  ",fs->mnt);
		l = Kb2str(fs->free,l);
		*l = '/';
		l = Kb2str(fs->blocks,l + 1);
		mvaddstr(num,0,line);
		fs = fs->next;
		dbug("fs@%d:%s\n",num,line);
		++num;
	}
done:
	if (unshown) {
		sysstat_t *u = s->next;
		int len = 7;
		size_t maxlen = sizeof(line) - 1;
		
		attrset(nostress);
		if (Numcol < maxlen)
			maxlen = Numcol;
		++Numlines;
		mvaddstr(num, 0, "hosts: ");
		while (u != s) {
			char *d,*b;
			size_t l;

			b = strchr(u->Hostinfo,' ');
			d = strchr(u->Hostinfo,'.');
			if (b && d) {
				if (b < d)
					l = b - u->Hostinfo;
				else
					l = d - u->Hostinfo;
			} else if (b) {
				l = b - u->Hostinfo;
			} else if (d) {
				l = d - u->Hostinfo;
			} else {
				assert(0);
			}
			len += l;
			if (len >= maxlen)
				break;
			strncpy(line,u->Hostinfo,l);
			line[l] = 0;
			if (now - u->TS > 60) 
				attrset(stress);
			addstr(line);
			attrset(nostress);
			u = u->next;
			if (u != s) {
				len += 2;
				if (len >= maxlen)
					break;
				addstr(", ");
			}
		}
		if (len < Numcol) {
			memset(line,' ',Numcol - len);
			line[Numcol - len] = 0;
			addstr(line);
		}
	} else 
		--num;
	memset(line,' ',Numcol);
	line[Numcol] = 0;
	while (++num < Numlines)
		mvaddstr(num,0,line);
	move(0,0);
	refresh();
	return 1;
}

// vim:tw=0
